#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
import calendar
import csv
import datetime
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import gzip
import os
import io
import logging
import re
import smtplib
import socket
import tempfile

import GeoIP
import six

PARSE_LINE = re.compile('^\[(.*?)\s.*?\]\s.*?\s(.*?)\s\".*?\"\s200\s\"(.*?)\s')

logger = logging.getLogger()


class MasterNodeInfo(object):
    def __init__(self, ip=None, country=None):
        self.ip = ip
        self.country = country


class StopLogsProcessing(Exception):
    pass


def get_logs(logs_dir, from_date):
    logger.info("Searching logs for analysis in %s", logs_dir)
    for f in os.listdir(logs_dir):
        path = os.path.join(logs_dir, f)
        if os.path.isfile(path):
            mod_time = datetime.datetime.fromtimestamp(os.path.getmtime(path))
            if from_date <= mod_time:
                yield path


def get_report_dates(params):
    year = params.year
    month = params.month
    _, last_day = calendar.monthrange(year, month)
    from_date = datetime.datetime(params.year, params.month, 1, hour=0, minute=0, second=0)
    to_date = datetime.datetime(params.year, params.month, last_day,
                                hour=23, minute=59, second=59)
    return from_date, to_date


def parse_log(log, mns_info, to_date):
    if log.endswith('.gz'):
        o = gzip.open
    else:
        o = open

    with o(log) as f:
        logger.debug("Parsing log %s", log)
        for line in f.readlines():
            parse_log_line(line, mns_info, to_date)
        logger.debug("Log %s parsed", log)


def parse_log_line(line, mns_info, to_date):
    m = re.match(PARSE_LINE, line)
    if m:
        dt = datetime.datetime.strptime(m.group(1), '%d/%b/%Y:%H:%M:%S')
        ip = m.group(2)
        mn_uid = m.group(3)
        if dt > to_date:
            raise StopLogsProcessing

        # Skip empty MN_UID header
        if len(mn_uid) == 1:
            return
        mns_info[mn_uid] = MasterNodeInfo(ip=ip)


def get_geoips(mns_info):
    logger.info("Fetching countries for master nodes ips")
    gi = GeoIP.new(GeoIP.GEOIP_MEMORY_CACHE)
    for mn_info in mns_info.itervalues():
        mn_info.country = gi.country_name_by_addr(mn_info.ip)
    logger.info("Countries fetched for master nodes ips")


def save_csv_report(mns_info, csv_file):
    logger.info("Saving CSV report")
    fieldnames = ['master_node_uid', 'country']
    writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
    writer.writeheader()

    for mn_uid, info in mns_info.iteritems():
        if info.country is None:
            logger.info("Undefined geo location for ip: %s", info.ip)
            continue
        writer.writerow({'master_node_uid': mn_uid,
                         'country': info.country.strip('\'"')})
    logger.info("CSV report saved")


def send_email(smtp_server, emails, csv_file, month_name, year):
    logger.info("Sending report")
    sender = 'robot@{0}'.format(socket.getfqdn())
    msg = MIMEMultipart()
    msg['Subject'] = 'IPs stats report on {0} {1}'.format(month_name, year)
    msg['From'] = sender
    msg['To'] = ', '.join(emails)

    csv_file.seek(io.SEEK_SET)
    attachment = MIMEText(csv_file.read())
    attachment.add_header('Content-Disposition', 'attachment',
                          filename=os.path.basename(csv_file.name))
    msg.attach(attachment)

    s = smtplib.SMTP(smtp_server)
    s.sendmail(sender, emails, msg.as_string())
    s.quit()
    logger.info("Report sent")


def send_report(params, mns_info):
    smtp_server = params.smtp_server
    month_name = calendar.month_name[params.month]
    emails = list(six.moves.map(six.text_type, params.emails.split(',')))
    year = params.year

    with tempfile.NamedTemporaryFile(
        prefix='countries_{0}_{1}_'.format(year, month_name),
        suffix='.csv'
    ) as csv_file:
        save_csv_report(mns_info, csv_file)
        send_email(smtp_server, emails, csv_file, month_name, year)


def execute(params):
    from_date, to_date = get_report_dates(params)
    logs = get_logs(params.logs_dir, from_date)
    sorted_logs = sorted(logs, key=lambda p: os.path.getmtime(p))
    mns_info = dict()
    for log in sorted_logs:
        try:
            parse_log(log, mns_info, to_date)
        except StopLogsProcessing:
            logger.debug("Log %s contains dates greater than %s. "
                  "Logs parsing finished", log, to_date)
    get_geoips(mns_info)
    send_report(params, mns_info)


def month_type(value):
    value = int(value)
    if value < 1:
        raise argparse.ArgumentTypeError("Minimum month is 1")
    if value > 12:
        raise argparse.ArgumentTypeError("Maximum month is 12")
    return value


def configure_logging():
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    ch.setFormatter(formatter)
    logger.addHandler(ch)


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        '-l', '--logs-dir',
        help="Path logs directory",
        default='/var/log/nginx'
    )
    parser.add_argument(
        '-m', '--month',
        help="Build a report for a given month. Current month is "
             "used by default",
        type=month_type,
        default=datetime.datetime.now().strftime('%m')
    )
    parser.add_argument(
        '-y', '--year',
        help="Build a report for a given year. Current year is "
             "used by default",
        type=int,
        default=datetime.datetime.now().strftime('%Y')
    )
    parser.add_argument(
        '-e', '--emails',
        help="Comma separated emails list",
        type=six.text_type
    )
    parser.add_argument(
        '-s', '--smtp-server',
        help="SMTP server",
        type=six.text_type,
        default='localhost'
    )
    params = parser.parse_args()
    configure_logging()

    try:
        execute(params)
    except:
        logger.exception("Logs parsing failed")
        raise


if __name__ == '__main__':
    main()
